/*
 * blueMail offline mail reader
 * area_header and area_list

 Copyright (c) 1996 Toth Istvan <stoty@vma.bme.hu>
 Copyright (c) 1998 William McBrine <wmcbrine@clark.net>
 Copyright (c) 1998 Ingo Brueckl <ib@wupperonline.de>

 Distributed under the GNU General Public License.
 For details, see the file COPYING in the parent directory. */

#include "bmail.h"

// ---------------------------------------------------------------------------
// Area header methods
// ---------------------------------------------------------------------------

area_header::area_header(bmail *bmA, int numA, const char *shortNameA,
                        const char *nameA, const char *descriptionA,
                        const char *areaTypeA, int typeA, int noOfLettersA,
                        int noOfPersonalA)
{
        bm = bmA;
        shortName = shortNameA;
        name = nameA;
        description = descriptionA;
        areaType = areaTypeA;
        type = typeA;
        num = numA;
        noOfLetters = noOfLettersA;
        noOfPersonal = noOfPersonalA;

        driver = bm->driverList->getDriver(num);
}

const char *area_header::getName() const
{
        return name;
}

const char *area_header::getShortName() const
{
        return shortName;
}

const char *area_header::getDescription() const
{
        return description;
}

const char *area_header::getAreaType() const
{
        return areaType;
}

int area_header::getNoOfLetters() const
{
        return noOfLetters;
}

int area_header::getNoOfUnread()
{
        return (isActive() ? (bm->driverList->getReadObject(driver))->
                getNoOfUnread(num - bm->driverList->getOffset(driver)) : 0);
}

int area_header::getNoOfPersonal() const
{
        return noOfPersonal;
}

int area_header::getUseAlias() const
{
        return type & ALIAS;
}

int area_header::isCollection() const
{
        return type & COLLECTION;
}

int area_header::isReplyArea() const
{
        return type & REPLYAREA;
}

int area_header::isActive() const
{
        return type & ACTIVE;
}

int area_header::isNetmail() const
{
        return (type & NETMAIL) && !(type & INTERNET);
}

// ---------------------------------------------------------------------------
// Arealist methods
// ---------------------------------------------------------------------------

area_list::area_list(bmail *bmA)
{
        bm = bmA;
        no = 0;

        int c = 0;
        do {
                no += bm->driverList->getDriver(no)->getNoOfAreas();
                c++;
        } while (c < bm->driverList->getNoOfDrivers());

        activeHeader = new int[no];
        areaHeader = new area_header *[no];

        specific_driver *actDriver;
        for (c = 0; c < no; c++) {
                actDriver = bm->driverList->getDriver(c);
                areaHeader[c] = actDriver->getNextArea();
        }

        current = shortlist = bm->resourceObject->isYes(LSAreaList);
        relist();

        // 1. Find out what types of areas we have (i.e. qwk, usenet... )
        // 2. Create the appropriate driver objects
        // 3. Find out the number of areas for each type
        // 4. Allocate the memory for the area_header descriptions
        // 5. Fill the area headers
}

area_list::~area_list()
{
        while (no)
                delete areaHeader[--no];
        delete[] areaHeader;
        delete[] activeHeader;
}

void area_list::relist()
{
        noActive = 0;
        shortlist = !shortlist;
        for (int c = 0; c < no; c++)
                if (areaHeader[c]->isActive() || !shortlist)
                        activeHeader[noActive++] = c;
}

const char *area_list::getShortName() const
{
        return areaHeader[current]->getShortName();
}

const char *area_list::getName() const
{
        return areaHeader[current]->getName();
}

const char *area_list::getName(int area)
{
        if ((area < 0) || (area >= no))
                fatalError("Internal error in area_list::getName.");
        return areaHeader[area]->getName();
}

const char *area_list::getDescription() const
{
        return areaHeader[current]->getDescription();
}

const char *area_list::getDescription(int area)
{
        if ((area < 0) || (area >= no))
                fatalError("Internal error in area_list::getDescription.");
        return areaHeader[area]->getDescription();
}

const char *area_list::getAreaType() const
{
        return areaHeader[current]->getAreaType();
}

int area_list::getNoOfLetters() const
{
        return areaHeader[current]->getNoOfLetters();
}

int area_list::getNoOfUnread() const
{
        return areaHeader[current]->getNoOfUnread();
}

int area_list::getNoOfPersonal() const
{
        return areaHeader[current]->getNoOfPersonal();
}

void area_list::getLetterList()
{
        bm->letterList = new letter_list(bm, current, isCollection() &&
                                        !isReplyArea());
}

int area_list::noOfAreas() const
{
        return no;
}

int area_list::noOfActive() const
{
        return noActive;
}

void area_list::gotoArea(int currentA)
{
        if ((currentA >= 0) && (currentA < no))
                current = currentA;
}

void area_list::gotoActive(int activeA)
{
        if ((activeA >= 0) && (activeA < noActive))
                current = activeHeader[activeA];
}

int area_list::getAreaNo() const
{
        return current;
}

int area_list::getActive()
{
        int c;

        for (c = 0; c < noActive; c++)
                if (activeHeader[c] >= current)
                        break;
        return c;
}

void area_list::enterLetter(int areaNo, char *from, char *to, char *subject,
                        int replyTo, int privat, net_address *netAddress,
                        char *filename, int length)
{
        net_address dummy_address;

        if (!netAddress) {
                dummy_address.zone = 0;
                netAddress = &dummy_address;
        }

        reply_driver *replyDriver = bm->driverList->getReplyDriver();

        gotoArea(areaNo);
        letter_header *newLetter = new letter_header(bm, subject, to, from,
                        "", replyTo, 0, 0, areaNo, privat, 0, replyDriver,
                        netAddress);

        replyDriver->enterLetter(newLetter, filename, length);

        delete newLetter;
        refreshArea();
}

void area_list::killLetter(int letterNo)
{
        bm->driverList->getReplyDriver()->killLetter(letterNo);
        refreshArea();
}

void area_list::makeReply()
{
        bm->driverList->getReplyDriver()->makeReply();
}

void area_list::refreshArea()
{
        delete areaHeader[REPLY_AREA];

        areaHeader[REPLY_AREA] =
                bm->driverList->getReplyDriver()->refreshArea();
        if (current == REPLY_AREA)
                bm->letterList->rrefresh();
}

int area_list::getUseAlias() const
{
        return areaHeader[current]->getUseAlias();
}

int area_list::isCollection() const
{
        return areaHeader[current]->isCollection();
}

int area_list::isReplyArea() const
{
        return areaHeader[current]->isReplyArea();
}

int area_list::isNetmail() const
{
        return areaHeader[current]->isNetmail();
}

int area_list::findNetmail() const
{
        int c;

        for (c = 0; c < no; c++)
                if (areaHeader[c]->isNetmail())
                        break;

        return (c < no) ? c : -1;
}
