/*
 * blueMail offline mail reader
 * letter_header and letter_list

 Copyright (c) 1996 Toth Istvan <stoty@vma.bme.hu>
 Copyright (c) 1998 William McBrine <wmcbrine@clark.net>
 Copyright (c) 1998 Ingo Brueckl <ib@wupperonline.de>

 Distributed under the GNU General Public License.
 For details, see the file COPYING in the parent directory. */

#include "bmail.h"

// ---------------------------------------------------------------------------
// Letter header methods
// ---------------------------------------------------------------------------

letter_header::letter_header(bmail *bmA, const char *subjectA,
                const char *toA, const char *fromA, const char *dateA,
                int replyToA, int LetterIDA, int msgNumA, int AreaIDA,
                int privatA, int lengthA, specific_driver *driverA,
                net_address *netAddrA)
{
        dl = bmA->driverList;
        ro = bmA->resourceObject;
        driver = driverA;
        readO = dl->getReadObject(driver);
        subject = strdupplus(subjectA);
        to = strdupplus(toA);
        from = strdupplus(fromA);
        date = strdupplus(dateA);
        replyTo = replyToA;
        LetterID = LetterIDA;
        AreaID = AreaIDA;
        privat = privatA;
        length = lengthA;
        msgNum = msgNumA;
        if (netAddrA)
                netAddr = *netAddrA;
        else
                netAddr.zone = 0;
}

letter_header::~letter_header()
{
        delete[] subject;
        delete[] to;
        delete[] from;
        delete[] date;
}

const char *letter_header::getSubject() const
{
        return subject;
}

const char *letter_header::getTo() const
{
        return to;
}

const char *letter_header::getFrom() const
{
        return from;
}

const char *letter_header::getDate() const
{
        return date;
}

int letter_header::getReplyTo() const
{
        return replyTo;
}

int letter_header::getLetterID() const
{
        return LetterID;
}

int letter_header::getAreaID() const
{
        return AreaID + dl->getOffset(driver);
}

int letter_header::getPrivate() const
{
        return privat;
}

const char *letter_header::getBody()
{
        return driver->getBody(AreaID, LetterID);
}

int letter_header::getLength() const
{
        return length;
}

specific_driver *letter_header::getDriver() const
{
        return driver;
}

int letter_header::getRead()
{
        return readO->getRead(AreaID, LetterID);
}

void letter_header::setRead()
{
        readO->setRead(AreaID, LetterID, 1);
}

const XTI_REC *letter_header::getStatus ()
{
  return readO->getStatus(AreaID, LetterID);
}

void letter_header::setStatus (tBYTE flags, tBYTE marks)
{
  readO->setStatus(AreaID, LetterID, flags, marks);
}

int letter_header::getMsgNum() const
{
        return msgNum;
}

net_address *letter_header::getNetAddr()
{
        return &netAddr;
}

bool letter_header::isPersonal ()
{
  const char *l, *a;

  l = ro->get(LoginName);
  a = ro->get(AliasName);

  return (strcasecmp(l, to) == 0) || (strcasecmp(a, to) == 0);
}

// ---------------------------------------------------------------------------
// Letterlist methods
// ---------------------------------------------------------------------------

letter_list::letter_list(bmail *bmA, int areaNumberA, int isCollA)
{
        dl = bmA->driverList;
        ro = bmA->resourceObject;
        areaNumber = areaNumberA;
        driver = dl->getDriver(areaNumber);
        areaNumber -= dl->getOffset(driver);
        readO = dl->getReadObject(driver);
        isColl = isCollA;
        init();
}

void letter_list::init()
{
        driver->selectArea(areaNumber);
        noOfLetters = driver->getNoOfLetters();

        activeHeader = new int[noOfLetters];
        letterHeader = new letter_header *[noOfLetters];

        driver->resetLetters();
        for (int c = 0; c < noOfLetters; c++)
                letterHeader[c] = driver->getNextLetter();
        currentLetter = shortlist = ro->isYes(LSLetterList);
        if ((noOfLetters > 1) && !isColl)
                lettqsort(0, noOfLetters - 1);
                //sort(LS_SUBJ);        // test/default sort key should be here

        relist();
}

letter_list::~letter_list()
{
        cleanup();
}

void letter_list::relist()
{
        noActive = 0;
        while (noOfLetters && !noActive) {
                shortlist = !shortlist;
                for (int c = 0; c < noOfLetters; c++)
                        if (!shortlist || !letterHeader[c]->getRead())
                                activeHeader[noActive++] = c;
        }
}

void letter_list::cleanup()
{
        while (noOfLetters)
                delete letterHeader[--noOfLetters];
        delete[] letterHeader;
        delete[] activeHeader;
}

void letter_list::lettqsort(int left, int right)
{
        letter_header *pivot, *lhtmp;
        const char *p;
        int lower, upper, d, t, p2;

        lower = left;
        upper = right;
        pivot = letterHeader[((left + right) >> 1)];
        p = pivot->getSubject();
        p2 = pivot->getLetterID();
        do {
                do {
                        d = strcasecmp(letterHeader[lower]->getSubject(), p);
                        t = (d < 0) || (!d && ((letterHeader[lower]->
                              getLetterID()) < p2));
                        if (t)
                                lower++;
                } while (t);
                do {
                        d = strcasecmp(letterHeader[upper]->getSubject(), p);
                        t = (d > 0) || (!d && ((letterHeader[upper]->
                              getLetterID()) > p2));
                        if (t)
                                upper--;
                } while (t);
                if (lower <= upper) {
                        if (lower != upper) {
                                lhtmp = letterHeader[lower];
                                letterHeader[lower] = letterHeader[upper];
                                letterHeader[upper] = lhtmp;
                        }
                        lower++;
                        upper--;
                }
        } while (lower < upper);
        if (left < upper)
                lettqsort(left, upper);
        if (lower < right)
                lettqsort(lower, right);
}

/*
void letter_list::sort(int keyord)      // lame, not really impl'd etc...
{
        switch (keyord) {
        case LS_SUBJ:
                lettqsort(0, noOfLetters - 1);
        }
}
*/

int letter_list::noOfLetter() const
{
        return noOfLetters;
}

int letter_list::noOfActive() const
{
        return noActive;
}

const char *letter_list::getSubject() const
{
        return letterHeader[currentLetter]->getSubject();
}

const char *letter_list::getFrom() const
{
        return letterHeader[currentLetter]->getFrom();
}

int letter_list::getMsgNum() const
{
        return letterHeader[currentLetter]->getMsgNum();
}

const char *letter_list::getTo() const
{
        return letterHeader[currentLetter]->getTo();
}

const char *letter_list::getDate() const
{
        return letterHeader[currentLetter]->getDate();
}

int letter_list::getReplyTo() const
{
        return letterHeader[currentLetter]->getReplyTo();
}

int letter_list::getAreaID() const
{
        return letterHeader[currentLetter]->getAreaID();
}

int letter_list::getPrivate() const
{
        return letterHeader[currentLetter]->getPrivate();
}

const char *letter_list::getBody()
{
        return (isColl ? driver->getBody(areaNumber, currentLetter) :
                letterHeader[currentLetter]->getBody());
}

int letter_list::getLength() const
{
        return letterHeader[currentLetter]->getLength();
}

net_address *letter_list::getNetAddr()
{
        return letterHeader[currentLetter]->getNetAddr();
}

int letter_list::getRead()
{
        return letterHeader[currentLetter]->getRead();
}

void letter_list::setRead()
{
        if (isColl)
                readO->setRead(areaNumber, currentLetter, 1);
        letterHeader[currentLetter]->setRead();
}

const XTI_REC *letter_list::getStatus()
{
        return letterHeader[currentLetter]->getStatus();
}

void letter_list::setStatus(tBYTE flags, tBYTE marks)
{
        if (isColl)
                readO->setStatus(areaNumber, currentLetter, flags, marks);
        letterHeader[currentLetter]->setStatus(flags, marks);
}

int letter_list::getCurrent() const
{
        return currentLetter;
}

int letter_list::getActive() const
{
        int c;

        for (c = 0; c < noActive; c++)
                if (activeHeader[c] >= currentLetter)
                        break;
        return c;
}

void letter_list::gotoLetter(int newLetter)
{
        if ((newLetter >= 0) && (newLetter < noOfLetters))
                currentLetter = newLetter;
}

void letter_list::gotoActive(int activeA)
{
        if ((activeA >= 0) && (activeA < noActive))
                currentLetter = activeHeader[activeA];
}

void letter_list::rrefresh()
{
        cleanup();
        init();
}

bool letter_list::isPersonal ()
{
  return letterHeader[currentLetter]->isPersonal();
}
