/*
 * blueMail offline mail reader
 * resource class

 Copyright (c) 1996 Toth Istvan <stoty@vma.bme.hu>
 Copyright (c) 1998 William McBrine <wmcbrine@clark.net>
 Copyright (c) 1998 Ingo Brueckl <ib@wupperonline.de>

 Distributed under the GNU General Public License.
 For details, see the file COPYING in the parent directory. */

#include <sys/stat.h>
#include <dirent.h>

#include "bmail.h"
#include "../interfac/error.h"

extern ErrorType error;

resource::resource ()
{
  int c;
  bool ok;
  FILE *configFile;
  char configFileName[MYMAXPATH];
  const char *envhome;
  const char *greeting = "\nWelcome new user!\n\n"
                         "A new (empty) " RCNAME " has been created. "
                         "If you continue now, " BM_NAME " will\n"
                         "use the default values. If you wish to edit your "
                         RCNAME " first, say 'N' at\n"
                         "the prompt.\n\nContinue? (y/n) ";

  for (c = 0; c < maxOptions; c++) resourceData[c].value = NULL;

  envhome = getenv("BMAIL");
  if (!envhome) envhome = getenv("HOME");
  if (!envhome) envhome = error.getOrigDir();

  set(homeDir, fixPath(envhome));
  mkfname(configFileName, get(homeDir), RCNAME);

  if (!(configFile = fopen(configFileName, "rt")))
  {
    if (createNewConfig(configFileName))
    {
      printf(greeting);
      c = getchar();
      if (toupper(c) == 'N') exit(1);
      else configFile = fopen(configFileName, "rt");
    }
    else fatalError("Could not open " RCNAME);
  }

  initinit();
  parseConfig(configFile);
  fclose(configFile);

  bmHomeInit();
  ok = verifyPaths();
  if (ok) ok = taglineCheck();
  if (!ok) fatalError("Unable to access data directories.");

  workinit();
}

resource::~resource ()
{
  clearDirectory(resourceData[PacketWorkDir].value);
  clearDirectory(resourceData[ReplyWorkDir].value);
  mychdir("..");
  rmdir(resourceData[PacketWorkDir].value);
  rmdir(resourceData[ReplyWorkDir].value);
  mychdir("..");
  rmdir(resourceData[WorkDir].value);

  for (int c = 0; c < maxOptions; c++) delete[] resourceData[c].value;
}

// for consistency, no path should end in a slash
const char *resource::fixPath (const char *path)
{
  size_t l;
  static char tmp[MYMAXPATH];

  if ((l = strlen(path)) > 0)
  {
    char d = path[l - 1];

    if (d == '/' || d == '\\')
    {
      mkfname(tmp, path, ".");
      return tmp;
    }
  }

  return path;
}

bool resource::checkPath (const char *onepath, bool show)
{
  DIR *dir;

  if ((dir = opendir(onepath)) == NULL)
  {
    if (show) printf("Creating %s...\n", onepath);
    if (mkdir(onepath, S_IRWXU) != 0) return false;
  }
  else closedir(dir);

  return true;
}

// create tagline file if it doesn't exist
bool resource::taglineCheck ()
{

#include "taglines.h"     // default taglines

  FILE *tagf;
  const char **p;
  bool succ = false;

  if ((tagf = fopen(resourceData[TaglineFile].value, "rt"))) succ = true;
  else
  {
    printf("Creating %s...\n", resourceData[TaglineFile].value);
    if ((tagf = fopen(resourceData[TaglineFile].value, "wt")))
    {
      for (p = defaultTaglines; *p; p++)
      {
        fputs(*p, tagf);
        fputc('\n', tagf);
      }
    succ = true;
    }
  }

  if (succ) fclose(tagf);

  return succ;
}

bool resource::verifyPaths ()
{
  if (checkPath(resourceData[bmailDir].value, true))
    if (checkPath(resourceData[PacketDir].value, true))
      if (checkPath(resourceData[ReplyDir].value, true))
        if (checkPath(resourceData[SaveDir].value, true))
          return true;
  return false;
}

void resource::createConfigBlock (FILE *F, const int from, const int to,
                                  const char *section)
{
  fprintf(F, "\n[%s]\n", section);
  for (int i = from; i <= to; i++) fprintf(F, "#%s:\n", OptionNames[i]);
}

bool resource::createNewConfig (const char *configname)
{

#include "bmailrc.h"

  FILE *bmrc;
  const char **p;

  printf("Creating %s...\n", configname);

  if ((bmrc = fopen(configname, "wt")))
  {
    for (p = bmailrcHeader; *p; p++)
    {
      fputs(*p, bmrc);
      fputc('\n', bmrc);
    }
    createConfigBlock(bmrc, dirFIRST, dirLAST, "Directories");
    createConfigBlock(bmrc, stringFIRST, stringLAST, "Strings");
    createConfigBlock(bmrc, cmdFIRST, cmdLAST, "Commands");
    createConfigBlock(bmrc, fileFIRST, fileLAST, "Files");
    createConfigBlock(bmrc, miscFIRST, miscLAST, "Miscellaneous");
    createConfigBlock(bmrc, colorFIRST, colorLAST, "Colors");
    fclose(bmrc);
    return true;
  }
  else return false;
}

void resource::parseConfig (FILE *configFile)
{
  char buffer[MYMAXLINE], tmp[50], *optName, *optValue, *pos;

  while (fgets(buffer, sizeof(buffer), configFile))
  {
    if (buffer[0] != '#' && buffer[0] != '[' && buffer[0] != '\n')
    {
      if ((pos = strchr(buffer, '\n')) != NULL) *pos = '\0';
      pos = buffer;

      //leading spaces
      while (*pos && (*pos == ' ' || *pos == '\t')) pos++;

      //optName
      optName = pos;
      while (*pos && *pos != ':' && *pos != ' ' && *pos != '\t') pos++;

      if (*pos) *pos++ = '\0';

      //chars between strings
      while (*pos && (*pos == ' ' || *pos == '\t' || *pos == ':')) pos++;

      //optValue
      optValue = pos;
      cropesp(optValue);

      for (int i = 0; i <= maxOptions; i++)
      {
        if (i == maxOptions)
        {
          sprintf(tmp, "Unknow option '%.32s'.", optName);
          fatalError(tmp);
        }

        if (strcasecmp(OptionNames[i], optName) == 0)
        {
          if (!*optValue)
          {
            sprintf(tmp, "Empty option '%.32s'.", optName);
            fatalError(tmp);
          }
          set(i, (i >= dirFIRST && i <= dirLAST) ? fixPath(optValue) : optValue);
          break;
        }
      }
    }
  }
}

const char *resource::get (int ID) const
{
  return ((ID < 0 || ID >= maxOptions) ? NULL : resourceData[ID].value);
}

void resource::set (int ID, const char *newValue)
{
  if (ID >= 0 && ID < maxOptions)
  {
    delete[] resourceData[ID].value;
    resourceData[ID].value = strdupplus(newValue);
  }
}

bool resource::isYes (int ID)
{
  if (ID < 0 || ID >= maxOptions || resourceData[ID].value == NULL)
    return false;
  else return (toupper(*resourceData[ID].value) == 'Y');
}

void resource::bmEachInit (int index, const char *dirname)
{
  char tmp[MYMAXPATH];

  if (!resourceData[index].value)
  {
    mkfname(tmp, resourceData[bmailDir].value, dirname);
    set(index, tmp);
  }
}

void resource::subWorkPath (int index, const char *dirname)
{
  char tmp[MYMAXPATH];

  mkfname(tmp, resourceData[WorkDir].value, dirname);
  set(index, tmp);
  if (!checkPath(tmp, false)) fatalError("Could not create work subdir.");
}

void resource::initinit ()
{
  char *p;

  set(arjUncompressCommand, DEFUNARJ);
#ifdef __MSDOS__
  set(arjCompressCommand, DEFARJ);
#endif
  set(zipUncompressCommand, DEFUNZIP);
  set(zipCompressCommand, DEFZIP);
  set(lhaUncompressCommand, DEFUNLHA);
  set(lhaCompressCommand, DEFLHA);

  p = getenv("EDITOR");
  set(editor, (p ? p : DEFEDIT));
}

void resource::workinit ()
{
  char tmp[MYMAXPATH];

  if (!resourceData[WorkDir].value) mytmpnam(tmp);
  else
  {
    mkfname(tmp, resourceData[WorkDir].value, "bmXXXXXX");
    if (!mktemp(tmp)) fatalError("Out of temporary filenames.");
  }
  set(WorkDir, tmp);
  checkPath(tmp, false);
  subWorkPath(PacketWorkDir, "downwork");
  subWorkPath(ReplyWorkDir, "upwork");
}

void resource::bmHomeInit ()
{
  char tmp[MYMAXPATH];

  if (!resourceData[bmailDir].value)
  {
    mkfname(tmp, resourceData[homeDir].value, "bmail");
    set(bmailDir, tmp);
  }

  bmEachInit(PacketDir, "down");
  bmEachInit(ReplyDir, "up");
  bmEachInit(SaveDir, "save");
  bmEachInit(AddressFile, ADDRBOOK);
  bmEachInit(TaglineFile, "taglines");
}
