/*
 * blueMail offline mail reader
 * Interface, ShadowedWin, ListWindow

 Copyright (c) 1996 Kolossvary Tamas <thomas@tvnet.hu>
 Copyright (c) 1997 John Zero <john@graphisoft.hu>
 Copyright (c) 1998 William McBrine <wmcbrine@clark.net>
 Copyright (c) 1998 Ingo Brueckl <ib@wupperonline.de>

 Distributed under the GNU General Public License.
 For details, see the file COPYING in the parent directory. */

#include "interfac.h"
#include "isoconv.h"
#include "mysystem.h"

Win::Win (int l, int c, int y, int x, chtype backg)
{
  win = newwin(l, c, y, x);
  Clear(backg);

  keypad(win, TRUE);
  cursor_off();
}

Win::~Win ()
{
  delwin(win);
}

void Win::Clear (chtype backg)
{
  wbkgdset(win, (backg & A_CHARTEXT) ? backg : backg | ' ');
  werase(win);
  wbkgdset(win, ' ');
  wattrset(win, backg);
}

void Win::putch (int y, int x, chtype z)
{
  mvwaddch(win, y, x, z);
}

void Win::putchnstr (int y, int x, const chtype *z, int len)
{
  mvwaddchnstr(win, y, x, (chtype *) z, len);
}

void Win::putstring (int y, int x, const char *z)
{
  mvwaddstr(win, y, x, (char *) z);
}

void Win::putchr (int y, int x, const char c)
{
  static char tmp[2] = {'\0', '\0'};

  tmp[0] = c;
  mvwaddstr(win, y, x, tmp);
}

void Win::attrib (chtype z)
{
  wattrset(win, z);
}

void Win::horizline (int y, int len)
{
  wmove(win, y, 1);
  whline(win, ACS_HLINE, len);
}

void Win::delay_update ()
{
  wnoutrefresh(win);
}

void Win::update ()
{
  wrefresh(win);
}

void Win::touch ()
{
  touchwin(win);
  wnoutrefresh(win);
}

void Win::wscroll (int i)
{
  scrollok(win, TRUE);
  wscrl(win, i);
  scrollok(win, FALSE);
}

void Win::cursor_on ()
{
  leaveok(win, FALSE);
#ifdef __PDCURSES__
  PDC_cursor_on();
#endif
}

void Win::cursor_off ()
{
  leaveok(win, TRUE);
#ifdef __PDCURSES__
  PDC_cursor_off();
#endif
}

int Win::keypressed ()
{
  nodelay(win, TRUE);
  return (wgetch(win) != ERR);
}

int Win::inkey ()
{
  nodelay(win, FALSE);
  return wgetch(win);
}

void Win::boxtitle (chtype backg, const char *title, chtype titleAttrib)
{
  wattrset(win, backg);
  box(win, 0, 0);

  if (title)
  {
    putch(0, 2, ACS_RTEE);
    putch(0, 3 + strlen(title), ACS_LTEE);
    wattrset(win, titleAttrib);
    putstring(0, 3, title);
    wattrset(win, backg);
  }
}

void Win::Border ()
{
  box(win, 0, 0);
}

ShadowedWin::ShadowedWin (int l, int c, int y, int x, chtype backg,
                          const char *title, chtype titleAttrib)
           : Win(l, c, y, x, backg)
{
  // The text to be in "shadow" is taken from different windows,
  // depending on the curses implementation. Note that the default,
  // "stdscr", just makes the shadowed area solid black; only with
  // ncurses and PDCurses does it draw proper shadows.

#ifdef USE_SHADOWS

  int i, j, sx, sy;
  chtype *right, *lower;

  #ifndef NCURSES_VERSION
    #ifdef __PDCURSES__
      WINDOW *newscr = curscr;
    #else
      WINDOW *newscr = stdscr;
    #endif
  #endif

  int limit = 2 + (interface->active() != letter);
  sx = c;
  sy = l;
  while ((sx + x) > (COLS - limit)) sx--;
  while ((sy + y) > (LINES - limit)) sy--;

  right = new chtype[(sy - 1) << 1];
  lower = new chtype[sx + 1];

  // gather the old text and attribute info:

  for (i = 0; i < (sy - 1); i++)
    for (j = 0; j <= 1; j++)
      right[(i << 1) + j] = mvwinch(newscr, (y + i + 1), (x + sx + j))
                          & (A_CHARTEXT | A_ALTCHARSET);

  mvwinchnstr(newscr, y + sy, x + 2, lower, sx);

  // redraw it in darkened form:

  shadow = newwin(sy, sx, y + 1, x + 2);
  leaveok(shadow, TRUE);
  for (i = 0; i < (sy - 1); i++)
    for (j = 0; j <= 1; j++)
      mvwaddch(shadow, i, sx - 2 + j, right[(i << 1) + j] | C_SHADOW);

  for (i = 0; i < sx; i++)
    mvwaddch(shadow, sy - 1, i, lower[i] & (A_CHARTEXT | A_ALTCHARSET)
                                         | C_SHADOW);
#endif

  boxtitle(backg, title, titleAttrib);

#ifdef USE_SHADOWS
  delete[] lower;
  delete[] right;
#endif
}

ShadowedWin::~ShadowedWin ()
{
#ifdef USE_SHADOWS
  delwin(shadow);
#endif
}

void ShadowedWin::delay_update ()
{
#ifdef USE_SHADOWS
  wnoutrefresh(shadow);
#endif
  Win::delay_update();
}

void ShadowedWin::touch ()
{
#ifdef USE_SHADOWS
  touchwin(shadow);
  wnoutrefresh(shadow);
#endif
  Win::touch();
}

int ShadowedWin::getstring (int y, int x, char *string, int n,
                            chtype inp_color, chtype norm_color)
{
  int i, j, end, c;
  char *tmp = new char[n];
  bool first_key = true;

  cropesp(string);

  wattrset(win, inp_color);
  for (i = 0; i < n; i++)
  {
    putch(y, x + i, ' ');
    tmp[i] = '\0';
  }
  putstring(y, x, string);

  cursor_on();
  update();

  i = end = 0;
  while (!end)
  {
    c = inkey();

    // those keys make the string "accepted" and editable
    if (first_key &&
        (c == KEY_LEFT || c == KEY_RIGHT || // c == KEY2_LEFT || c == KEY2_RIGHT
         c == 127 || c == KEY_DC || c == 8 || c == KEY_BACKSPACE ||
         c == KEY_HOME || c == KEY_END))
    {
      strcpy(tmp, string);
      i = strlen(string);
    }

    first_key = false;

    switch (c)
    {
      case KEY_DOWN:
        end++;

      case KEY_UP:
        end++;

      case '\r':
      case '\n':
        end++;

      case ESC_KEY:
        end++;
        break;

      case KEY_LEFT:
      // ?? case KEY2_LEFT:
        if (i > 0) i--;
        break;

      case KEY_RIGHT:
      // ?? case KEY2_RIGHT:
        if ((i < n - 1) && tmp[i]) i++;
        break;

      case 127:     // Delete key
      case KEY_DC:
        strncpy(&tmp[i], &tmp[i + 1], n - i - 1);
        tmp[n - 1] = '\0';
        break;

      case 8:     // Ctrl-H
      case KEY_BACKSPACE:
        if (i > 0)
        {
          strncpy(&tmp[i - 1], &tmp[i], n - i - 1);
          tmp[n - 1] = '\0';
          i--;
        }
        break;

      case KEY_HOME:
        i = 0;
        break;

      case KEY_END:
        while ((i < n - 1) && tmp[i]) i++;
        break;

      default:
        for (j = n - 1; j > i; j--) tmp[j] = tmp[j - 1];
        tmp[i] = c;
        if (i < n - 1) i++;
    }

    for (j = 0; j < n; j++) putchr(y, x + j, (tmp[j] ? tmp[j] : ' '));
    wmove(win, y, x + i);
    update();
  }

  if (tmp[0]) strcpy(string, tmp);
  wattrset(win, norm_color);
  j = strlen(string);
  for (i = 0; i < n; i++) putchr(y, x + i, ((i < j) ? string[i] : ' '));

  update();
  cursor_off();

  delete[] tmp;
  return end - 1;
}

InfoWin::InfoWin (int l, int c, int y, int x, chtype backg, const char *title,
                  chtype titleAttrib, int bottx, int topx)
       : ShadowedWin(l, c, y, x, backg, title, titleAttrib)
{
  lineBuf = new char[COLS];
  info = new Win(l - bottx, c - 2, y + topx, x + 1, backg);
}

InfoWin::~InfoWin ()
{
  delete info;
  delete[] lineBuf;
}

void InfoWin::delay_update (int complete)
{
  if (complete) ShadowedWin::delay_update();
  info->delay_update();
}

void InfoWin::touch (int complete)
{
  if (complete) ShadowedWin::touch();
  info->touch();
}

void InfoWin::oneline (int i, chtype ch)
{
  charconv_in(lineBuf);
  info->attrib(ch);
  info->putstring(i, 0, lineBuf);
}

void InfoWin::Scroll (int i)
{
  info->wscroll(i);
}

ListWindow::ListWindow ()
{
  position = active = 0;
  oldPos = oldActive = -100;
}

void ListWindow::checkPos (int limit)
{
  if (active >= limit) active = limit - 1;

  if (active < 0) active = 0;

  if (active < position) position = active;
  else
    if (active - position >= list_max_y) position = active - list_max_y + 1;

  if (limit > list_max_y)
  {
    if (position < 0) position = 0;
    else if (position > limit - list_max_y) position = limit - list_max_y;
  }
  else position = 0;
}

void ListWindow::Draw ()
{
  int i, j, limit = NumOfItems();
  const info = 0;

  checkPos(limit);
  i = position - oldPos;

  switch (i)
  {
    case -1:
    case 1:
      list->Scroll(i);
      // no break, must go through default

    case 0:
      if (active != oldActive)
      {
        j = oldActive - position;
        if (j >= 0 && j < list_max_y) oneLine(j);   // de-select old line
      }
      oneLine(active - position);                   // select new line
      list->delay_update(info);
      break;

    default:
      for (j = 0; j < list_max_y; j++)
      {
        oneLine(j);
        if (position + j == limit - 1) j = list_max_y;
      }
      list->touch(info);
  }
  oldPos = position;
  oldActive = active;
}

// highlight, if drawing the active bar
void ListWindow::DrawOne (int i, chtype color)
{
#ifdef __FIXCURSES__
  list->oneline(i, interface->colorlist.colormode(color,
                   ((position + i == active) ? A_STANDOUT : A_NORMAL)));
#else
  list->oneline(i, ((position + i == active) ? A_STANDOUT : color));
#endif
}

void ListWindow::Move (direction dir)
{
  int limit = NumOfItems();

  switch (dir)
  {
    case UP:
      active--;
      break;

    case DOWN:
      active++;
      break;

    case PGUP:
      position -= list_max_y;
      active -= list_max_y;
      break;

    case PGDN:
      position += list_max_y;
      active += list_max_y;
      break;

    case HOME:
      active = 0;
      break;

    case END:
      active = limit - 1;
      break;
  }
  checkPos(limit);
}

void ListWindow::KeyHandle (int key)
{
  bool draw = true;

  switch (key)
  {
    case KEY_DOWN:
      Move(DOWN);
      break;

    case KEY_UP:
      Move(UP);
      break;

    case KEY_HOME:
      Move(HOME);
      break;

    case KEY_END:
      Move(END);
      break;

    case KEY_PPAGE:
    //case KEY2_PPAGE:
      Move(PGUP);
      break;

    case ' ':
    case KEY_NPAGE:
    //case KEY2_NPAGE:
      Move(PGDN);
      break;

    default:
      draw = false;
      extrakeys(key);
  }
  if (draw) Draw();
}

Interface::Interface ()
{
  prevstate = state = nostate;
  isoconvert =
#if defined(__MSDOS__) || defined(__EMX__)
               false;
#else
               true;
#endif
}

Interface::~Interface ()
{
  delete screen;
  touchwin(stdscr);
  refresh();
  endwin();
}

void Interface::init_colors ()
{
  for (int b = COLOR_BLACK; b <= COLOR_WHITE; b++)
    for (int f = COLOR_BLACK; f <= COLOR_WHITE; f++)
      init_pair((f << 3) + b, f, b);

  // Color pair 0 (BLACK on BLACK) cannot be set, it defaults
  // to the color the terminal had before libcurses started.
  // So we use WHITE on WHITE for BLACK on BLACK which is needed
  // for the window's shadows. As a result, WHITE on WHITE can't
  // be used. :-(

  init_pair((COLOR_WHITE << 3) + COLOR_WHITE, COLOR_BLACK, COLOR_BLACK);
}

void Interface::screen_init ()
{
  initscr();
  refresh();
  start_color();
  init_colors();
  cbreak();
  noecho();
  nonl();

  // make a new background window, fill it with ACS_BOARD characters
  screen = new Win(0, 0, 0, 0, C_MBACK | ACS_BOARD);

  if (COLS < 60 || LINES < 20)
    fatalError("A screen at least 60x20 is required.");

  // border and title
  static char tmp[80];
  sprintf(tmp, BM_TOPHEADER, BM_NAME, BM_MAJOR, BM_MINOR);
  screen->boxtitle(C_MBORDER, tmp, C_MBACK | A_BOLD);

  // help window area
  screen->attrib(C_MSEPBOTT);
  screen->horizline(LINES - 5, COLS - 2);

  screen->delay_update();
}

int Interface::WarningWindow (const char *warning, const char **selectors,
                              int items)
{
  static const char *yesno[] = {"Yes", "No"};

  int x, y, z, itemlen, c, curitem, def_val = 0;
  bool result = false;

  if (!selectors) selectors = yesno;     // default options

  // calculate the window width and maximum item width:

  for (itemlen = 0, curitem = 0; curitem < items; curitem++)
  {
    z = strlen(selectors[curitem]);
    if (z > itemlen) itemlen = z;
  }
  itemlen += 2;
  x = itemlen * (items + 1);
  z = strlen(warning);
  if ((z + 4) > x) x = z + 4;

  ShadowedWin warn(7, x, (LINES >> 1) - 4, (COLS - x) >> 1, C_WARNTEXT);
  warn.putstring(2, (x - z) >> 1, warning);

  y = x / (items + 1);

  // display each item:

  const char *p;
  for (curitem = 0; curitem < items; curitem++)
  {
    z = strlen(selectors[curitem]) >> 1;
    x = (curitem + 1) * y - z;
    p = selectors[curitem] + 1;

    warn.attrib(C_WARNTEXT);
    warn.putstring(4, x + 1, p--);

    warn.attrib(C_WARNKEYS);
    warn.putch(4, x, *p);
  }
  warn.delay_update();

  // main loop -- highlight selected item and process keystrokes:

  while (!result)
  {
    for (curitem = 0; curitem < items; curitem++)
    {
      z = strlen(selectors[curitem]) >> 1;
      x = (curitem + 1) * y - z;
      warn.putch(4, x - 1, (def_val == curitem) ? '[' : ' ');
      warn.putch(4, x + strlen(selectors[curitem]),
                 (def_val == curitem) ? ']' : ' ');
    }
    warn.update();

    c = warn.inkey();

    for (curitem = 0; curitem < items && !result; curitem++)
    {
      p = selectors[curitem];
      if (c == tolower(*p) || c == toupper(*p))
      {
        def_val = curitem;
        result = true;
      }
    }

    if (!result)
      switch (c)
      {
        case ESC_KEY:
          def_val = items;     // will force return-value to be -1
          result = true;
          break;

        case KEY_LEFT:
          if (def_val == 0) def_val = items;
          def_val--;
          break;

        case KEY_RIGHT:
        case TAB_KEY:
          if (++def_val == items) def_val = 0;
          break;

        case '\r':
        case '\n':
          result = true;
      }
  }

  return (items - 1) - def_val;     // the order is inverted
}

void Interface::nonFatalError (const char *warn)
{
  static const char *ok[] = {"Ok"};

  WarningWindow(warn, ok, 1);
  changestate(state);
}

statetype Interface::active ()
{
  return state;
}

statetype Interface::prevactive ()
{
  return prevstate;
}

void Interface::destroystate (statetype o)
{
  helpwindow.Delete();

  switch (o)
  {
    case nostate:
      break;

    case packetlist:
      packets.Delete();
      break;

    case arealist:
      areas.Delete();
      break;

    case letterlist:
      letters.Delete();
      break;

    case address:
      addresses.Delete();
      break;

    case tagwin:
      taglines.Delete();
      break;

    case littlearealist:
      delete littleareas;
      break;

    case letter_help:
    case letter:
      letterwindow.Delete();
      break;

    case ansiwin:
      ansiwindow.Delete();
      break;
  }
  screen->touch();
}

void Interface::newstate (statetype n)
{
  if (state != n)
  {
    prevstate = state;
    state = n;
  }

  switch (n)
  {
    case nostate:
      break;

    case packetlist:
      currList = &packets;
      packets.MakeActive();
      break;

    case arealist:
      currList = &areas;
      areas.MakeActive();
      break;

    case letterlist:
      currList = &letters;
      letters.MakeActive();
      break;

    case letter:
      letterwindow.MakeActive(prevstate == letterlist);   // display starting
      break;                                              // from line 1

    case letter_help:
      letterwindow.MakeActive(false);
      break;

    case littlearealist:
      letterwindow.MakeActive(false);
      littleareas = new LittleAreaListWindow(Key);
      currList = littleareas;
      break;

    case address:
      currList = &addresses;
      addresses.MakeActive();
      KeyHandle();
      break;

    case tagwin:
      currList = &taglines;
      taglines.MakeActive();
      break;

    case ansiwin:
      ansiwindow.MakeActive();
  }
  helpwindow.MakeActive();
}

void Interface::changestate (statetype n)
{
  destroystate(state);
  newstate(n);
}

int Interface::select ()
{
  switch (state)
  {
    case packetlist:
      unsaved_reply = any_read = false;
      switch (packets.OpenPacket())
      {
        case PKT_UNCOMP_FAILED:
          nonFatalError("Could not uncompress packet");
          break;

        case PKT_TYPE_UNKNOWN:
          nonFatalError("Packet type not recognized");
          break;

        default:
          changestate(arealist);
          areas.FirstUnread();
      }
      break;

    case arealist:
      areas.Select();
      if (bm.areaList->getNoOfLetters() > 0)
      {
        bm.areaList->getLetterList();
        changestate(letterlist);
        letters.FirstUnread();
        if (bm.resourceObject->isYes(SkipLetterList)) select();
      }
      break;

    case letterlist:
      letters.Select();
      changestate(letter);
      any_read = true;
      break;

    case letter:
      letterwindow.Next();
      break;

    case letter_help:
      back();
      break;

    case littlearealist:
      changestate(letter);
      if (bm.areaList->isNetmail())
        if ((Key == 'N') || (Key == 'E')) newstate(address);
        else
        {
          net_address nm = letterwindow.PickNetAddr();
          letterwindow.set_Letter_Params(&nm, NULL);
        }
      Key = '\0';
      letterwindow.EnterLetter();
      changestate(letter);
      break;

    case address:
    case tagwin:
      currList->KeyHandle('\n');
      return back();

    default:
      ;
  }

  return 0;
}

int Interface::back ()
{
  int r = 0;

  switch (state)
  {
    case packetlist:
      if (WarningWindow("Do you really want to quit?") > 0)
      {
        destroystate(state);
        return 1;                // program ends now
      }
      else changestate(packetlist);
      break;

    case arealist:
      if (any_read)
      {
        if (bm.resourceObject->isYes(SaveLastreadPointers)) bm.saveRead();
        else if ((r = WarningWindow("Save lastread pointers?")) > 0)
          bm.saveRead();

        if (r == -1)
        {
          changestate(arealist);
          break;
        }
        else any_read = false;
      }

      if (unsaved_reply)
        if ((r = WarningWindow("The content of the reply area has changed. "
           "Create new packet?")) > 0) create_reply_packet();
      if (r == -1) changestate(arealist);
      else
      {
        bm.Delete();
        changestate(packetlist);
      }
      break;

    case letterlist:
      delete bm.letterList;     // ???
      changestate(arealist);
      break;

    case littlearealist:
    case letter_help:
      changestate(letter);
      break;

    case letter:
      changestate(letterlist);
      break;

    case address:
    case tagwin:
      changestate(prevstate);
      return 1;

    case ansiwin:
      changestate(prevstate);
      break;

    default:
      ;
  }

  return 0;
}

#ifdef SIGWINCH
void Interface::sigwinch ()
{
  destroystate(state);

  delete screen;
  endwin();
  screen_init();

  newstate(state);
  doupdate();
}
#endif

void Interface::kill_letter ()
{
  if (WarningWindow("Are you sure you want to delete this letter?") > 0)
  {
    bm.areaList->killLetter(bm.letterList->getMsgNum());
    unsaved_reply = true;
    changestate(letterlist);
    if (!bm.areaList->getNoOfLetters()) back();
  }
  else changestate(state);
}

void Interface::create_reply_packet ()
{
  bm.areaList->makeReply();
  unsaved_reply = false;
}

void Interface::ansiLoop (const char *source, const char *title)
{
  ansiwindow.set(source, title);
  changestate(ansiwin);
}

void Interface::TransToggle ()
{
  isoconvert = !isoconvert;
  // the following is a dirty trick, but prevstate must not be letterlist
  // to prevent letter's MakeActive from redrawing it from line 1 and setting
  // the read flag
  if (state == letter) prevstate = littlearealist;
  changestate(state);
}

void Interface::setUnsaved ()
{
  unsaved_reply = true;
}

void Interface::setAnyRead ()
{
  any_read = true;
}

void Interface::KeyHandle ()     // main loop
{
  int end = 0;

  while (!end)                   // all statetypes (windows) loop here
  {                              // this is a good time to perform all the
    doupdate();                  // (delayed) updates
    Key = screen->inkey();

    if (Key >= 'a' && Key <= 'z') Key = toupper(Key);

    switch (Key)
    {
      case 'Q':
      case ESC_KEY:
      case KEY_BACKSPACE:
        end = back();
        break;

      case '\n':
      case '\r':
        end = select();
        break;

      default:
        switch (state)
        {
          case letter:
            letterwindow.KeyHandle(Key);
            break;

          case letter_help:
            back();
            break;

          case ansiwin:
            ansiwindow.KeyHandle(Key);
            break;

          default:
            currList->KeyHandle(Key);
        }
    }
  }
}

int Interface::midpos (int nocols, int max_l, int max_m, int max_r)
{
  return max_l + 1 + ((COLS - nocols - max_l - 1 - max_r - 1 - max_m) >> 1);
}

int Interface::endpos (int nocols, int max_r)
{
  return COLS - nocols - 1 - max_r;
}
